/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef PAPYRUSGROUP_H
#define PAPYRUSGROUP_H

#include <list>
#include <map>
#include <vector>

#include <papyrus/drawable.h>

/**
 * \page guide_group Groups
 *
 * A group is a container for ordered collections of shapes, with the shapes
 * at the top (front of the container) drawn last and the shapes at the
 * bottom drawn first (end of the container).
 *
 * The current implementation is very simplistic, and doesn't even try to
 * do anything smart like taking into account occlusion or anything else. Instead
 * all objects are simply drawn and it is left to cairo to take care of
 * efficiency. At the current time there are no plans to improve this behavior,
 * since it runs 'fast enough' for now.
 *
 * Groups are also useful since transforms upon the group are also applied
 * to the contained objects. Thus, scaling, translating or rotating a group
 * has the same effect as performing those operations upon each member of
 * the group.
 *
 * The following contains an example that adds several shapes to a group,
 * then rotates the entire group by 45 degrees counter-clockwise.
 *
 * \image html group.png "Group and shapes drawn by example code below"
 *
 * \dontinclude factory.cpp
 * \skip example_group
 * \until }
 *
 */

namespace Papyrus
{

  /**
   * A group is a container for ordered collections of shapes, with the shapes
   * at the top (front of the container) drawn last and the shapes at the
   * bottom drawn first (end of the container).
   *
   * The current implementation is very simplistic, and doesn't even try to
   * do anything smart like taking into account occlusion or anything else. Instead
   * all objects are simply drawn and it is left to cairo to take care of
   * efficiency. At the current time there are no plans to improve this behavior,
   * since it runs 'fast enough' for now.
   *
   * Groups are also useful since transforms upon the group are also applied
   * to the contained objects. Thus, scaling, translating or rotating a group
   * has the same effect as performing those operations upon each member of
   * the group.
   *
   * \image html group.png "Group contain an ordered set of shapes"
   *
   *
   * @author Rick L Vinyard Jr
   */
  class Group : public Drawable
  {
    public:
      typedef PapyrusSmartPointer<Group> pointer;

      Group();

      static pointer create();

      typedef std::list<Drawable::pointer> Children;

      friend class Canvas;

      virtual ~Group();

      virtual bool add( Drawable::pointer item, bool drawing_only=false );

      virtual bool remove( Drawable::pointer object );
      
      /**
       * The main purpose of this method is so that a child may use the
       * this pointer to remove itself on destruction.
       */
      virtual bool remove( Drawable* object );
      
      virtual bool clear();

      virtual bool raise( Drawable::pointer item );
      virtual bool raise_to_top( Drawable::pointer item );
      virtual bool lower( Drawable::pointer item );
      virtual bool lower_to_bottom( Drawable::pointer item );

      Children& get_children();

      /**
       * Overrides parent method for determining whether (x,y) is inside the drawable.
       * If the group itself is selectable, performs a simple extents test.
       * If the group is not selectable, performs a test on each of the children, returning
       * true if any of the children are selectable.
       */
      virtual bool inside( double x, double y );

      virtual std::vector<Drawable::pointer> select( double x, double y, unsigned depth = 1 );

      sigc::signal<void, Drawable::pointer>& signal_child_added();

      sigc::signal<void, Drawable::pointer>& signal_child_removed();

      virtual bool is_group();

      PAPYRUS_CLASS_NAME("Group");

      PAPYRUS_CLONE_METHOD( Group );

    protected:

      // maintaining both items as list for drawing order and connections for quick access
      Children m_children;
      typedef std::map<Drawable::pointer, sigc::connection> Connections;
      Connections m_redraw_connections;
      Connections m_changed_connections;

      sigc::signal<void, Drawable::pointer> m_signal_child_added;
      sigc::signal<void, Drawable::pointer> m_signal_child_removed;

      /**
       * Virtual method called when a child is changed.
       * Parameters indicate which child changed and contain hint as to which aspect
       * of the child changed.
       *
       * When a child is added, the child's changed signal is actually connected to the
       * private proxy method, which in turn calls the virtual method. This allows children
       * to implement their own behavior by reimplementing this method without worrying about
       * modifying the add() or remove() methods.
       */
      virtual void on_child_changed( Drawable::pointer child );

      /**
       * Virtual method that recalculates the extents and emits the changed signal.
       *
       * Children should reimplement this method if they need to perform
       * special behavior when an event occurs that could effect the
       * extents, and that behavior needs to take place before the
       * signal is emitted.
       */
      virtual void update_extents();

      void on_child_need_redraw( double x, double y, double w, double h, Drawable::pointer child );
      virtual void draw( Cairo::RefPtr<Cairo::Context> cairo );

      void rebuild_extents();

  };

}

#endif
